#!/usr/bin/env python3
#
#  This file is part of ProfileView.
#
#  ProfileView is free software: you can redistribute it and/or modify
#  it under the terms of the CeCILL 2.1 Licence
#
#  ProfileView is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
#  You should have received a copy of the Licence CeCILL 2.1 along
#  with ProfileView. If not, see <https://cecill.info/>.
#

import sys, errno, os
import argparse
import math
from collections import defaultdict

import numpy as np
from ete3 import Tree

from pv_utils import *


def getBestSubtreeThreshold( mod_name, seq_scores, seq_names, seq_subtree ):
    thr = [0.0,0.0]
    points = [ seq_scores[i] for i,s in enumerate(seq_names) if s not in seq_subtree ]
    return np.amax(points,axis=0) if len(points) > 0 else thr


def getCentroid(points):
    a = np.asarray(points)
    num = a.shape[0]
    return [ np.sum(a[:,0])/num, np.sum(a[:,1])/num ] if num > 0 else [0.0,0.0]


def getRepresentedSequences( mod_name, seq_scores, seq_names, seq_subtree ):
    thr = getBestSubtreeThreshold( mod_name, seq_scores, seq_names, seq_subtree )
    indices = np.nonzero(np.any(np.greater(seq_scores,thr),axis=1))[0]
    scores = [ seq_scores[i] for i in indices ]
    names = uniq( seq_names[i] for i in indices )
    ctr = getCentroid(scores)
    ctr_norm = math.hypot(*ctr) if ctr is not None else 0.0
    return { "model":mod_name, "covered":names, "num_sequences":len(seq_subtree), "covered_score":ctr_norm, "thr":thr }


def getBestRepresentativeModel( best, a ):
    best_covered = len(best["covered"])
    best_score = best["covered_score"]
    a_covered = len(a["covered"])
    a_score = a["covered_score"]
    if (a_covered > best_covered) or (a_covered == best_covered and a_score > best_score):
        best = a
    return best


def hasBetterCoverage(a,b):
    a_covered = len(a["covered"])
    b_covered = len(b["covered"])
    return a_covered > b_covered



def main( argv = None ):

    # GET PARAMETERS
    parser = argparse.ArgumentParser()
    parser.add_argument('-t', '--tree',   dest='treeFile', required=True, help='Input tree (in newick format) obtained from the hierarchical clustering')
    parser.add_argument('-s', '--scores', dest='scoreFile', required=True, help='Score file obtained by the script parseHMMER.py')
    parser.add_argument('-m', '--models', dest='modelFile', help='List of model identifiers to consider')
    parser.add_argument('--sorted',       dest='isScoreSorted', action='store_true', help='Use this flag if score file is sorted by score (descending order)')
    parser.add_argument('--min-coverage', dest='minCoverage', type=float, default=0.5)
    parser.add_argument('-o', '--out',    dest='outFile',   default='out.reprTree', help='output prefix')
    opts = parser.parse_args()

    # VALIDATE PARAMETERS
    validParameters = True
    if not os.path.isfile(opts.treeFile):
        validParameters = False
        print_error("input tree file \"" + opts.treeFile + "\" does not exist.")
    if not os.path.isfile(opts.scoreFile):
        validParameters = False
        print_error("score file \"" + opts.scoreFile + "\" does not exist.")
    if not validParameters:
        return 1

    # LOAD MODEL FILE (IF PROVIDED)
    modSet = set()
    if opts.modelFile is not None:
        with open(opts.modelFile,'r') as fh:
            for line in fh:
                modid = line.strip()
                if modid != "":
                    modSet.add(modid)

    # LOAD FUNCTION TREE
    funTree = Tree(opts.treeFile)
    seqSet = set([ n.name for n in funTree.iter_leaves() ])

    # LOAD SCORE FILE
    modSeqs = defaultdict(list)
    modScores = defaultdict(list)
    with open(opts.scoreFile,'r') as fh:
        for line in fh:
            if not line.startswith('#'):
                fields = line.strip().split('\t') # hmm_name, hmm_len, seq_name, seq_len, seq_family, seq_func, bitscore, mean-bitscore, ...
                mod_name = fields[0]
                seq_name = fields[2]
                if opts.modelFile is None: # no model file given, cosider the set of models mentioned in the score file
                    modSet.add(mod_name)
                if seq_name in seqSet:
                    modSeqs[mod_name].append(seq_name)
                    modScores[mod_name].append([float(fields[x]) for x in [7,11]]) # [ bitscore, mean-bitscore ]

    seqDict = defaultdict(set)
    for mod_name in modScores: # in modDict
        if mod_name in modSet: # and (len(modDict[mod_name]) > 0):
            for i in np.argmax(modScores[mod_name],axis=0):
                seq_name = modSeqs[mod_name][i]
                seqDict[seq_name].add(mod_name)

    # (POST-ORDER) VISIT OF THE FUNCTIONAL TREE TO FIND REPRESENTATIVE MODELS
    nameIdx=1
    for node in funTree.iter_descendants("postorder"):
        if node.is_leaf(): # do not find best representative for leaves
            node.add_feature("models",seqDict[node.name])
            node.add_feature("sequences",set([node.name]))
            node.add_feature("num_covered",0)
            node.add_feature("mod_support",0.0)
            #print( 'LEAF({}): {}'.format(node.name,len(node.models)) )
            continue
        # internal node (different from the root)
        assert len(node.children) == 2
        node.name = 'I{}'.format(nameIdx)
        nameIdx += 1
        # gather models and sequences from the children
        node.add_feature("models", node.children[0].models | node.children[1].models)
        node.add_feature("sequences", node.children[0].sequences | node.children[1].sequences)
        #print( 'INTERNAL({}): {}'.format(node.name,node.models) )
        # test whether any of the model is representative for the sequence set
        best = { "model":"", "covered":[], "num_sequences":len(node.sequences), "covered_score":0.0, "thr":[0.0,0.0] }
        best_list = []
        good_list  = []
        for mod_name in node.models:
            rs = getRepresentedSequences(mod_name, modScores[mod_name], modSeqs[mod_name], node.sequences)
            rs_covered = len(rs["covered"])
            rs_support = float(rs_covered)/rs["num_sequences"]
            best_covered = len(best["covered"])
            if rs_covered > best_covered:
                best_list = [ (rs["model"],rs["covered_score"]) ]
            elif rs_covered == best_covered:
                best_list.append( (rs["model"],rs["covered_score"]) )
            if rs_support >= .9:
                good_list.append( (rs["model"],rs["covered_score"]) )
            best = getBestRepresentativeModel(best,rs)
        best_list.sort(key=lambda x:float(x[1]),reverse=True)
        good_list.sort(key=lambda x:float(x[1]),reverse=True)
        node.add_feature("repr_model",best["model"])
        node.add_feature("best_models",best_list)
        node.add_feature("good_models",good_list)
        node.add_feature("covered",best["covered"])
        node.add_feature("num_covered",len(node.covered))
        node.add_feature("num_sequences",len(node.sequences))
        node.add_feature("mod_support",float(node.num_covered)/node.num_sequences)
        node.add_feature("covered_score",best["covered_score"])
        node.add_feature("thr",best["thr"])
        node.support = node.mod_support if (node.mod_support >= opts.minCoverage and node.num_covered != node.children[0].num_covered and node.num_covered != node.children[1].num_covered) else 0.0
        node.support = node.support if len(node.sequences) > 2 or (node.num_covered == 2 and node.num_sequences == 2) else 0.0
        node.add_feature("conf",node.support)

    with open('{}.models.tsv'.format(opts.outFile),'w') as ofh:
        fmt = ['{nodeid}','{repr_model}','{mod_support:.2f}','{num_covered}','{num_sequences}','{seq_name}','{is_covered}','{best_models}','{good_models}']
        for node in funTree.iter_descendants('preorder'):
            if node.is_leaf() or node.num_covered <= 2:
                continue
            if node.support > 0.0: # has a representative model
                covered = set(node.covered)
                best_set = set()
                good_set = set()
                best_list=[]
                for p in node.best_models:
                    if p[0] not in best_set:
                        best_list.append(p[0])
                        best_set.add(p[0])
                good_list=[]
                for p in node.good_models:
                    if p[0] not in best_set and p[0] not in good_set:
                        good_list.append(p[0])
                        good_set.add(p[0])
                for seq_name in node.sequences:
                    ofh.write('\t'.join(fmt).format( nodeid=node.name, repr_model=node.repr_model, mod_support=node.mod_support,
                        num_covered=node.num_covered, num_sequences=node.num_sequences, seq_name=seq_name,
                        is_covered=(seq_name in covered), best_models=';'.join(best_list), good_models=';'.join(good_list) ))
                    ofh.write('\n')


    funTree.write(features=['name','conf','repr_model','num_covered'], format=2, outfile='{}.nhx'.format(opts.outFile))
    return 0


# Check if the program is not being imported
if __name__ == "__main__":
    sys.exit(main())

