#!/usr/bin/env python3
#
#  This file is part of ProfileView.
#
#  ProfileView is free software: you can redistribute it and/or modify
#  it under the terms of the CeCILL 2.1 Licence
#
#  ProfileView is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
#  You should have received a copy of the Licence CeCILL 2.1 along
#  with ProfileView. If not, see <https://cecill.info/>.
#

import sys, os
import argparse

from operator import itemgetter
from pv_utils import *

def main( argv = None ):

    # GET PARAMETERS
    parser = argparse.ArgumentParser()
    parser.add_argument('--seq-list',     dest='seqListFile', required=True, help='CSV file containing, for each line, sequence-name, seq-function, seq-family, seq-length')
    parser.add_argument('--hmm-list',     dest='hmmListFile', required=True, help='CSV file containing, for each line, hmm-name, pfam-domain, nseq, model-length')
    parser.add_argument('--scores',       dest='scoreFile',   required=True, help='score file obtained with parse_tblout_v2.py from the output obtained with hmmsearch')
    parser.add_argument('-n', '--nseq',   dest='nseq',        type=int, default=20, help='consider only HMMs with a number of sequences greater or equal than this value (default:5)')
    parser.add_argument('-k', '--k-best', dest='kBest',       type=int, default=1, help='consider only the k best models mapped against each sequence (default:1)')
    parser.add_argument('-p', '--prefix', dest='outPrefix',   default='out', help='output file prefix')
    args = parser.parse_args()

    # VALIDATE PARAMETERS
    validParameters = True
    if not os.path.isfile(args.seqListFile):
        validParameters = False
        print_error("sequence list file \"{}\" does not exist.".format(args.seqListFile))
    if not os.path.isfile(args.hmmListFile):
        validParameters = False
        print_error("HMM list file \"{}\" does not exist.".format(args.hmmListFile))
    if not os.path.isfile(args.scoreFile):
        validParameters = False
        print_error("score file \"{}\" does not exist.".format(args.scoreFile))
    if not validParameters:
        return 1

    # HANDLE "BAD" PARAMETERS
    if args.kBest < 1:
        print_warning("-k/--k-best parameter must be a positive integer, default value of 1 will be used")
        args.kBest = 1

    # LOAD SEQUENCE DICTIONARY
    seqDict = {}
    with open(args.seqListFile,'r') as seqListFile:
        for line in seqListFile:
            record = [ x.strip() for x in line.split(',') ]
            seq_name, seq_func, seq_fam = record[0:3]
            seq_len = int(record[3])
            seqDict[seq_name] = { 'Function':seq_func, 'Family':seq_fam, 'Length':seq_len }

    # LOAD HMMER MODEL DICTIONARY
    hmmDict = {}
    with open(args.hmmListFile,'r') as hmmListFile:
        for line in hmmListFile:
            record = [ x.strip() for x in line.split(',') ]
            hmm_name, pfam_acc = record[0:2]
            hmm_nseq = int(record[2])
            hmm_len  = int(record[3])
            if int(hmm_nseq) >= args.nseq:
                hmmDict[hmm_name] = { 'Domain':pfam_acc, 'NSEQ':hmm_nseq, 'Length':hmm_len }

    print_status("sequences: {}".format(len(seqDict)))
    print_status("models with >= {} sequences: {}".format(args.nseq,len(hmmDict)))

    allHits = []
    strictHits = []
    hmmStats = {}
    seqStats = {}
    with open(args.scoreFile,'r') as scoreFile:
        for line in scoreFile:
            line = line.strip()
            if (not line) or line.startswith('#'):
                continue
            # the score file is formatted with the following (tab-separated) fields:
            # hmm_name  hmm_len  seq_name  seq_len  seq_family  seq_func  bitscore  mean_score  mcscore  mean_mcs  wcscore  mean_wcs  ident  hmm_cov  hit_type
            # 0         1        2         3        4           5         6         7           8        9         10       11        12     13       14
            fields   = [ x.strip() for x in line.split('\t') ]
            hmm_name = fields[0]
            hmm_len  = int(fields[1])
            seq_name = fields[2]
            seq_len  = int(fields[3])
            bitscore = float(fields[6])
            mean_bs  = float(fields[7])
            mcscore  = float(fields[8])
            mean_mcs = float(fields[9])
            wcscore  = float(fields[10])
            mean_wcs = float(fields[11])
            identity = float(fields[12])
            hmm_cov  = float(fields[13])
            hit_type = fields[14]

            if hmm_name in hmmDict:
                if hmm_name not in hmmStats:
                    hmmStats[hmm_name] = { 'FULL':0, 'OVER':0, 'PART':0 }
                if seq_name not in seqStats:
                    seqStats[seq_name] = { 'FULL':0, 'OVER':0, 'PART':0, 'MODELS':[] }
                hit = [ hmm_name, seq_name, bitscore, mean_bs, mcscore, mean_mcs, wcscore, mean_wcs, identity, hmm_cov, hit_type ]
                allHits.append(hit)
                seqStats[seq_name]['MODELS'].append( (hmm_name,mean_bs,mean_wcs,hmm_len,hmm_cov,hit_type) )
                if (hit_type == 'FULL' or
                        (hit_type == 'PART' and hmm_cov >= 75.0) or
                        (hit_type == 'OVER' and hmm_cov >= 30.0)):
                    strictHits.append(hit)
                if hit_type == 'FULL':
                    hmmStats[hmm_name]['FULL'] += 1
                    seqStats[seq_name]['FULL'] += 1
                elif hit_type == 'PART':
                    hmmStats[hmm_name]['PART'] += 1
                    seqStats[seq_name]['PART'] += 1
                elif hit_type == 'OVER':
                    hmmStats[hmm_name]['OVER'] += 1
                    seqStats[seq_name]['OVER'] += 1
    print_status("{} hits loaded ({} strict)".format(len(allHits),len(strictHits)))

    # Discard partial/incomplete sequences
    with open("{}.discarded".format(args.outPrefix),'w') as outFile:
        partialSequences = []
        for seq_name in seqStats:
            stats = seqStats[seq_name]
            tot  = stats['FULL'] + stats['PART'] + stats['OVER']
            full = float(stats['FULL'])/tot if tot > 0 else 0.0
            part = float(stats['PART'])/tot if tot > 0 else 0.0
            over = float(stats['OVER'])/tot if tot > 0 else 0.0
            if (full < .5 and part < .5) or over >= .3: # significant number of sequence-model "overlapping" hits (i.e., the sequence is likely to be incomplete)
                partialSequences.append(seq_name)
                outFile.write("{}\t{:.4f}\t{:.4f}\t{:.4f}\n".format(seq_name,full,over,part))
        for seq_name in partialSequences:
            del seqStats[seq_name]
            del seqDict[seq_name]
    print_status("{} putative incomplete sequences, {} remaining".format(len(partialSequences),len(seqDict)))

    kBest1 = set()
    kBest2 = set()
    with open("{}.seq.stats".format(args.outPrefix),'w') as ofh:
        for sname in seqStats:
            seqStats[sname]['MODELS'].sort(key=itemgetter(2),reverse=True) # sort by mean-hcscore
            for i in range(min(args.kBest,len(seqStats[sname]['MODELS']))):
                kBest1.add(seqStats[sname]['MODELS'][i][0])
            seqStats[sname]['MODELS'].sort(key=itemgetter(1),reverse=True) # sort by mean-bitscore
            for i in range(min(args.kBest,len(seqStats[sname]['MODELS']))):
                kBest2.add(seqStats[sname]['MODELS'][i][0])
                #kBestModelSet.add(seqStats[sname]['MODELS'][i][0])
                cols = [ sname, seqDict[sname]['Family'], seqDict[sname]['Function'] ] + [ str(x) for x in seqStats[sname]['MODELS'][i] ]
                ofh.write( '\t'.join(cols) + '\n' )
    kBestModelSet = kBest2 | kBest1
    print_status("kBest models: {}".format(len(kBestModelSet)))

    with open('{}.used_models.list'.format(args.outPrefix), 'w') as f:
        for m in kBestModelSet:
            f.write('{}\n'.format(m))

    # OUTPUT FEATURE MATRIX BASED ON ALL HITS
    bestMeanBSDict  = {}
    bestMeanHCSDict = {}
    for hit in allHits:
        key = tuple(hit[0:2]) # (hmm_name,seq_name)
        bitscore = float(hit[2])
        wcscore = float(hit[6])
        mcscore = float(hit[4])
        mean_bs, mean_cms, mean_wcs = [ float(hit[x]) for x in [3,5,7] ] # mean_bit_score mean_consensus-match_score mean_highly-conserved-position_score
        if key[0] not in kBestModelSet:
            continue
        if (key not in bestMeanBSDict or mean_bs > bestMeanBSDict[key]):
            bestMeanBSDict[key]  = mean_bs
        if (key not in bestMeanHCSDict or mean_wcs > bestMeanHCSDict[key]):
            bestMeanHCSDict[key] = mean_wcs  #mean_hcs # mean_mcs # mean_hcs
    print_status("kBest hits: {}".format(len(bestMeanBSDict)))

    with open("{}.feat".format(args.outPrefix),'w') as outFile:
        outFile.write( '\t'.join(['SeqName','Label','Family'] +
            [ name.split('_')[0]+'_MBS'  for name in kBestModelSet ] +
            [ name.split('_')[0]+'_MHCS' for name in kBestModelSet ]) + '\n' )
        for seq_name in seqDict:
            scores  = [ "{:.4f}".format(bestMeanBSDict.get((hmm_name,seq_name),0.0))  for hmm_name in kBestModelSet ] + \
                      [ "{:.4f}".format(bestMeanHCSDict.get((hmm_name,seq_name),0.0)) for hmm_name in kBestModelSet ]
            out_row = [ seq_name, seqDict[seq_name]['Function'], seqDict[seq_name]['Family'] ] + scores
            outFile.write( '\t'.join(out_row) + '\n' )

    # OUTPUT FEATURE MATRIX BASED ONLY ON STRICT HITS
#    bestMeanBSDict  = {}
#    bestMeanHCSDict = {}
#    for hit in strictHits:
#        key = tuple(hit[0:2]) # (hmm_name,seq_name)
#        mean_bs, mean_cms, mean_hcs = [ float(x) for x in hit[3:8:2] ]
#        if key[0] not in kBestModelSet:
#            continue
#        if (key not in bestMeanBSDict or mean_bs > bestMeanBSDict[key]):
#            bestMeanBSDict[key]  = mean_bs
#            bestMeanHCSDict[key] = mean_mcs 
#            bestMeanWSDict[key]  = mean_hcs
#    print_status("kBest (strict) hits: {}".format(len(bestMeanBSDict)))
#
#    with open("{}.strict.feat".format(args.outPrefix),'w') as outFile:
#        outFile.write( '\t'.join(['SeqName','Label','Family'] +
#            [ name.split('_')[0]+'_MBS'  for name in kBestModelSet ] +
#            [ name.split('_')[0]+'_MWS'  for name in kBestModelSet ] +
#            [ name.split('_')[0]+'_MHCS' for name in kBestModelSet ]) + '\n' )
#        for seq_name in seqDict:
#            scores = [ "{:.4f}".format(bestMeanBSDict.get((hmm_name,seq_name),0.0))  for hmm_name in kBestModelSet ] + \
#                     [ "{:.4f}".format(bestMeanWSDict.get((hmm_name,seq_name),0.0))  for hmm_name in kBestModelSet ] + \
#                     [ "{:.4f}".format(bestMeanHCSDict.get((hmm_name,seq_name),0.0)) for hmm_name in kBestModelSet ]
#            out_row = [ seq_name, seqDict[seq_name]['Function'], seqDict[seq_name]['Family'] ] + scores
#            outFile.write( '\t'.join(out_row) + '\n' )

    return 0

# Check if the program is not being imported
if __name__ == "__main__":
    sys.exit(main())

