#
#  This file is part of ProfileView.
#
#  ProfileView is free software: you can redistribute it and/or modify
#  it under the terms of the CeCILL 2.1 Licence
#
#  ProfileView is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
#  You should have received a copy of the Licence CeCILL 2.1 along
#  with ProfileView. If not, see <https://cecill.info/>.
#

import math
import operator

from BitVector import BitVector
from pv_utils  import *

class hhm():

    def __init__(self, hhm_name, aa2idx, neff_hmm, neff_m, consensus, master_seq, match_states):
        self.name         = hhm_name
        self.aa2idx       = aa2idx
        self.consensus    = consensus
        self.master_seq   = master_seq
        self.match_states = match_states
        self.neff_hmm     = neff_hmm
        self.neff_m       = neff_m
        self.conserved    = BitVector( bitstring = ''.join([ '1' if c.isupper() else '0' for c in self.consensus ]) )


    def __str__(self):
        return "NAME: {}\nCONSENSUS: {}\nCONSERVED: {}\nMASTERSEQ: {}".format(self.name, self.consensus, self.conserved, self.master_seq)

    def frequency(self, state, aa):
        aaidx = self.aa2idx[aa]
        val = self.match_states[state][aaidx]
        return 2**(-val/1000.0)

    def state_frequencies(self, state):
        out = {}
        for aa in self.aa2idx:
            aai = self.aa2idx[aa]
            val = self.match_states[state][aai]
            out[aa] = 2**(-val/1000.0)
        return out

    def size(self):
        return len(self.master_seq)
    
    def get_consensus(self):
        return self.consensus


def parse_hhm(fh):
    hhm_name     = ""
    aa_list      = []
    aa2idx       = {}
    consensus    = ""
    master_seq   = ""
    match_states = []
    neff_hmm     = 0.0
    neff_m       = [99.999]
    while True:
        line = fh.readline()
        if line == '':
            break
        tokens = line.split()
        if len(tokens) == 0:
            continue
        elif tokens[0] == "NAME":
            hhm_name = tokens[1]
        elif tokens[0] == "NEFF":
            neff_hmm = float(tokens[1])
        elif tokens[0] == "SEQ":
            while not line.startswith('>Consensus'):
                line = fh.readline()
            # found consensus sequence id
            line = fh.readline()
            while line and line[0] != '>':
                consensus += ''.join(line.split())
                line = fh.readline()
            # found master sequence id
            line = fh.readline()
            while line and line[0] != '>' and line[0] != '#':
                master_seq += ''.join(line.split())
                line = fh.readline()
            # consensus/master sequences processed, discard remaining sequences until the end of the SEQ block
            while line and line[0] != '#':
                line = fh.readline()
        elif tokens[0] == "NULL":
            null_values = [ int(v) if is_int(v) else 99999 for v in tokens[1:21] ]
            match_states.append(null_values)
        elif tokens[0] == "HMM":
            # load AA alphabet
            aa_list = tokens[1:21]
            for i,aa in enumerate(aa_list):
                aa2idx[aa] = i
            # skip first two lines (transitions definitions/start probabilities)
            fh.readline() # M->M M->I M->D I->M I->I D->M D->D Neff NeffI NeffD
            fh.readline() #    0    *    *    0    *    0    *    *     *     *
            line = fh.readline()
            while line and not line.startswith("//"):
                tokens = line.split()
                if len(tokens) == 23:
                    state_values = [ int(val) if is_int(val) else 99999 for val in tokens[2:22] ]
                    match_states.append(state_values)
                elif len(tokens) == 10:
                    val = tokens[7]
                    neff_m.append( int(val)/1000.0 if is_int(val) else 99.999 )
                line = fh.readline()
    assert len(match_states) == len(master_seq)+1
    return hhm( hhm_name, aa2idx, neff_hmm, neff_m, consensus, master_seq, match_states )


